from .abstract_ranking import RankingElection  # noqa

from .plurality import (  # noqa
    Plurality,
    SNTV,
)

from .borda import Borda  # noqa
from .stv import STV, IRV, SequentialRCV  # noqa
from .alaska import Alaska  # noqa
from .top_two import TopTwo  # noqa
from .dominating_sets import DominatingSets  # noqa
from .condo_borda import CondoBorda  # noqa
from .top_two import TopTwo  # noqa
from .plurality_veto import PluralityVeto  # noqa
from .random_dictator import RandomDictator  # noqa
from .boosted_random_dictator import BoostedRandomDictator  # noqa



from .comentions import comention, comention_above, comentions_matrix  # noqa
from .boost import boost_prob, boost_matrix  # noqa
from .candidate_distance import candidate_distance, candidate_distance_matrix  # noqa



from fractions import Fraction
from typing import Union, Sequence, Optional, TypeVar, cast, Literal
from itertools import permutations
import math
import random
from .ballot import Ballot
from .pref_profile import PreferenceProfile

COLOR_LIST = [
    "#0099cd",
    "#ffca5d",
    "#00cd99",
    "#99cd00",
    "#cd0099",
    "#9900cd",
    "#8dd3c7",
    "#bebada",
    "#fb8072",
    "#80b1d3",
    "#fdb462",
    "#b3de69",
    "#fccde5",
    "#bc80bd",
    "#ccebc5",
    "#ffed6f",
    "#ffffb3",
    "#a6cee3",
    "#1f78b4",
    "#b2df8a",
    "#33a02c",
    "#fb9a99",
    "#e31a1c",
    "#fdbf6f",
    "#ff7f00",
    "#cab2d6",
    "#6a3d9a",
    "#b15928",
    "#64ffda",
    "#00B8D4",
    "#A1887F",
    "#76FF03",
    "#DCE775",
    "#B388FF",
    "#FF80AB",
    "#D81B60",
    "#26A69A",
    "#FFEA00",
    "#6200EA",
]


def ballots_by_first_cand(profile: PreferenceProfile) -> dict[str, list[Ballot]]:
    """
    Partitions the profile by first place candidate. Assumes there are no ties within first place
    positions of ballots.

    Args:
        profile (PreferenceProfile): Profile to partititon.

    Returns:
        dict[str, list[Ballot]]:
            A dictionary whose keys are candidates and values are lists of ballots that
            have that candidate first.
    """
    cand_dict: dict[str, list[Ballot]] = {c: [] for c in profile.candidates}

    for b in profile.ballots:
        if not b.ranking:
            raise TypeError("Ballots must have rankings.")
        else:
            # find first place candidate, ensure there is only one
            first_cand = list(b.ranking[0])
            if len(first_cand) > 1:
                raise ValueError(f"Ballot {b} has a tie for first.")

            cand_dict[first_cand[0]].append(b)

    return cand_dict


COB = TypeVar("COB", PreferenceProfile, tuple[Ballot, ...], Ballot)


def remove_cand(
    removed: Union[str, list],
    profile_or_ballots: COB,
    condense: bool = True,
    leave_zero_weight_ballots: bool = False,
) -> COB:
    """
    Removes specified candidate(s) from profile, ballot, or list of ballots. When a candidate is
    removed from a ballot, lower ranked candidates are moved up.
    Automatically condenses any ballots that match as result of scrubbing.

    Args:
        removed (Union[str, list]): Candidate or list of candidates to be removed.
        profile_or_ballots (Union[PreferenceProfile, tuple[Ballot,...], Ballot]): Collection
            of ballots to remove candidates from.
        condense (bool, optional): Whether or not to return a condensed profile. Defaults to True.
        leave_zero_weight_ballots (bool, optional): Whether or not to leave ballots with zero
            weight in the PreferenceProfile. Defaults to False.

    Returns:
        Union[PreferenceProfile, tuple[Ballot,...],Ballot]:
            Updated collection of ballots with candidate(s) removed.
    """
    if isinstance(removed, str):
        removed = [removed]

    # map to tuple of ballots
    if isinstance(profile_or_ballots, PreferenceProfile):
        ballots = profile_or_ballots.ballots
    elif isinstance(profile_or_ballots, Ballot):
        ballots = (profile_or_ballots,)
    else:
        ballots = profile_or_ballots[:]

    scrubbed_ballots = [Ballot()] * len(ballots)
    for i, ballot in enumerate(ballots):
        new_ranking = []
        new_scores = {}
        if ballot.ranking:
            for s in ballot.ranking:
                new_s = []
                for c in s:
                    if c not in removed:
                        new_s.append(c)
                if len(new_s) > 0:
                    new_ranking.append(frozenset(new_s))

        if ballot.scores:
            new_scores = {
                c: score for c, score in ballot.scores.items() if c not in removed
            }

        if len(new_ranking) > 0 and len(new_scores) > 0:
            scrubbed_ballots[i] = Ballot(
                ranking=tuple(new_ranking), weight=ballot.weight, scores=new_scores
            )
        elif len(new_ranking) > 0:
            scrubbed_ballots[i] = Ballot(
                ranking=tuple(new_ranking), weight=ballot.weight
            )

        elif len(new_scores) > 0:
            scrubbed_ballots[i] = Ballot(weight=ballot.weight, scores=new_scores)

        # else ballot exhausted
        else:
            scrubbed_ballots[i] = Ballot(weight=Fraction(0))

    # return matching input data type
    if isinstance(profile_or_ballots, PreferenceProfile):
        clean_profile = PreferenceProfile(
            ballots=tuple([b for b in scrubbed_ballots if b.weight > 0]),
            candidates=tuple(
                [c for c in profile_or_ballots.candidates if c not in removed]
            ),
        )

        if leave_zero_weight_ballots:
            clean_profile = PreferenceProfile(
                ballots=tuple(scrubbed_ballots),
                candidates=tuple(
                    [c for c in profile_or_ballots.candidates if c not in removed]
                ),
            )

        if condense:
            clean_profile = clean_profile.condense_ballots()

        return cast(COB, clean_profile)

    elif isinstance(profile_or_ballots, Ballot):
        clean_profile = None

        if leave_zero_weight_ballots:
            clean_profile = PreferenceProfile(
                ballots=tuple(scrubbed_ballots),
            )
        else:
            clean_profile = PreferenceProfile(
                ballots=tuple([b for b in scrubbed_ballots if b.weight > 0]),
            )

        if condense:
            clean_profile = clean_profile.condense_ballots()

        return cast(COB, clean_profile.ballots[0])
    else:
        clean_profile = None

        if leave_zero_weight_ballots:
            clean_profile = PreferenceProfile(
                ballots=tuple(scrubbed_ballots),
            )
        else:
            clean_profile = PreferenceProfile(
                ballots=tuple([b for b in scrubbed_ballots if b.weight > 0]),
            )

        if condense:
            clean_profile = clean_profile.condense_ballots()

        return cast(COB, clean_profile.ballots)


def add_missing_cands(profile: PreferenceProfile) -> PreferenceProfile:
    """
    Add any candidates from `profile.candidates` that are not listed on a ballot
    as tied in last place. Automatically condenses profile.

    Args:
        profile (PreferenceProfile): Input profile.

    Returns:
        PreferenceProfile
    """

    new_ballots = [Ballot()] * len(profile.ballots)
    candidates = set(profile.candidates)

    for i, ballot in enumerate(profile.ballots):
        if not ballot.ranking:
            raise TypeError("Ballots must have rankings.")
        else:
            b_cands = [c for s in ballot.ranking for c in s]
            missing_cands = candidates.difference(b_cands)

            new_ranking = (
                list(ballot.ranking) + [missing_cands]
                if len(missing_cands) > 0
                else ballot.ranking
            )

            new_ballots[i] = Ballot(
                id=ballot.id,
                weight=ballot.weight,
                voter_set=ballot.voter_set,
                ranking=tuple([frozenset(s) for s in new_ranking]),
            )

    return PreferenceProfile(
        ballots=tuple(new_ballots), candidates=tuple(candidates)
    ).condense_ballots()


def validate_score_vector(score_vector: Sequence[Union[float, Fraction]]):
    """
    Validator function for score vectors. Vectors should be non-increasing and non-negative.

    Args:
        score_vector (Sequence[Union[float, Fraction]]): Score vector.

    Raises:
        ValueError: If any score is negative.
        ValueError: If score vector is increasing at any point.

    """

    for i, score in enumerate(score_vector):
        # if score is negative
        if score < 0:
            raise ValueError("Score vector must be non-negative.")

        if i > 0:
            # if the current score is bigger than prev
            if score > score_vector[i - 1]:
                raise ValueError("Score vector must be non-increasing.")


def score_profile_from_rankings(
    profile: PreferenceProfile,
    score_vector: Sequence[Union[float, Fraction]],
    to_float: bool = False,
    tie_convention: Literal["high", "average", "low"] = "low",
) -> Union[dict[str, Fraction], dict[str, float]]:
    """
    Score the candidates based on a score vector. For example, the vector (1,0,...) would
    return the first place votes for each candidate. Vectors should be non-increasing and
    non-negative. Vector should be as long as ``max_ballot_length`` in the profile.
    If it is shorter, we add 0s. Unlisted candidates receive 0 points.


    Args:
        profile (PreferenceProfile): Profile to score.
        score_vector (Sequence[Union[float, Fraction]]): Score vector. Should be
            non-increasing and non-negative. Vector should be as long as ``max_ballot_length`` in
            the profile. If it is shorter, we add 0s.
        to_float (bool, optional): If True, compute scores as floats instead of Fractions.
            Defaults to False.
        tie_convention (Literal["high", "average", "low"], optional): How to award points for
            tied rankings. Defaults to "low", where any candidates tied receive the lowest possible
            points for their position, eg three people tied for 3rd would each receive the points
            for 5th. "high" awards the highest possible points, so in the previous example, they
            would each receive the points for 3rd. "average" averages the points, so they would each
            receive the points for 4th place.

    Returns:
        Union[dict[str, Fraction], dict[str, float]]:
            Dictionary mapping candidates to scores.
    """
    validate_score_vector(score_vector)

    max_length = profile.max_ballot_length
    if len(score_vector) < max_length:
        score_vector = list(score_vector) + [0] * (max_length - len(score_vector))

    scores = {c: Fraction(0) for c in profile.candidates}
    for ballot in profile.ballots:
        current_ind = 0
        if not ballot.ranking:
            raise TypeError("Ballots must have rankings.")
        else:
            for s in ballot.ranking:
                position_size = len(s)
                if len(s) == 0:
                    raise TypeError(f"Ballot {ballot} has an empty ranking position.")
                local_score_vector = score_vector[
                    current_ind : current_ind + position_size
                ]

                if tie_convention == "high":
                    allocation = max(local_score_vector)
                elif tie_convention == "low":
                    allocation = min(local_score_vector)
                elif tie_convention == "average":
                    allocation = sum(local_score_vector) / position_size
                else:
                    raise ValueError(
                        (
                            "tie_convention must be one of 'high', 'low', 'average', "
                            f"not {tie_convention}"
                        )
                    )
                for c in s:
                    scores[c] += Fraction(allocation) * ballot.weight
                current_ind += position_size

    if to_float:
        return {c: float(v) for c, v in scores.items()}
    return scores


def first_place_votes(
    profile: PreferenceProfile,
    to_float: bool = False,
    tie_convention: Literal["high", "average", "low"] = "average",
) -> Union[dict[str, Fraction], dict[str, float]]:
    """
    Computes first place votes for all candidates in a ``PreferenceProfile``.

    Args:
        profile (PreferenceProfile): The profile to compute first place votes for.
        to_float (bool): If True, compute first place votes as floats instead of Fractions.
            Defaults to False.
        tie_convention (Literal["high", "average", "low"], optional): How to award points
            for tied first place votes. Defaults to "average", where if n candidates are tied for
            first, each receives 1/n points. "high" would award them each one point, and "low" 0.

    Returns:
        Union[dict[str, Fraction],dict[str, float]]:
            Dictionary mapping candidates to number of first place votes.
    """
    # equiv to score vector of (1,0,0,...)
    return score_profile_from_rankings(
        profile, [1] + [0] * (profile.max_ballot_length - 1), to_float, tie_convention
    )


def mentions(
    profile: PreferenceProfile,
    to_float: bool = False,
) -> Union[dict[str, Fraction], dict[str, float]]:
    """
    Calculates total mentions for a ``PreferenceProfile``.

    Args:
        profile (PreferenceProfile): PreferenceProfile of ballots.
        to_float (bool): If True, compute mention as floats instead of Fractions.
            Defaults to False.

    Returns:
        Union[dict[str, Fraction], dict[str, float]]:
            Dictionary mapping candidates to mention totals (values).
    """
    mentions = {c: Fraction(0) for c in profile.candidates}

    for ballot in profile.ballots:
        if not ballot.ranking:
            raise TypeError("Ballots must have rankings.")
        else:
            for s in ballot.ranking:
                for cand in s:
                    mentions[cand] += ballot.weight
    if to_float:
        return {c: float(v) for c, v in mentions.items()}
    return mentions


def borda_scores(
    profile: PreferenceProfile,
    borda_max: Optional[int] = None,
    to_float: bool = False,
    tie_convention: Literal["high", "average", "low"] = "low",
) -> Union[dict[str, Fraction], dict[str, float]]:
    r"""
    Calculates Borda scores for a ``PreferenceProfile``. The Borda vector is
    :math:`(n,n-1,\dots,1)` where :math:`n` is the ``borda_max`.

    Args:
        profile (PreferenceProfile): ``PreferenceProfile`` of ballots.
        borda_max (int, optional): The maximum value of the Borda vector. Defaults to
            the length of the longest allowable ballot in the profile.
        to_float (bool): If True, compute Borda as floats instead of Fractions.
            Defaults to False.
        tie_convention (Literal["high", "average", "low"], optional): How to award points for
            tied rankings. Defaults to "low", where any candidates tied receive the lowest possible
            points for their position, eg three people tied for 3rd would each receive the points
            for 5th. "high" awards the highest possible points, so in the previous example, they
            would each receive the points for 3rd. "average" averages the points, so they would each
            receive the points for 4th place.

    Returns:
        Union[dict[str, Fraction], dict[str, float]]:
            Dictionary mapping candidates to Borda scores.
    """
    if not borda_max:
        borda_max = profile.max_ballot_length

    score_vector = list(range(borda_max, 0, -1))

    return score_profile_from_rankings(profile, score_vector, to_float, tie_convention)


def tiebreak_set(
    r_set: frozenset[str],
    profile: Optional[PreferenceProfile] = None,
    tiebreak: str = "random",
    scoring_tie_convention: Literal["high", "average", "low"] = "low",
) -> tuple[frozenset[str], ...]:
    """
    Break a single set of candidates into multiple sets each with a single candidate according
    to a tiebreak rule. Rule 1: random. Rule 2: first-place votes; break the tie based on
    first-place votes in the profile. Rule 3: borda; break the tie based on Borda points in the
    profile.

    Args:
        r_set (frozenset[str]): Set of candidates on which to break tie.
        profile (PreferenceProfile, optional): Profile used to break ties in first-place votes or
            Borda setting. Defaults to None, which implies a random tiebreak.
        tiebreak (str, optional): Tiebreak method to use. Options are "random", "first_place", and
            "borda". Defaults to "random".
        scoring_tie_convention (Literal["high", "average", "low"], optional): How to award points
            for tied rankings. Defaults to "low", where any candidates tied receive the lowest
            possible points for their position, eg three people tied for 3rd would each receive the
            points for 5th. "high" awards the highest possible points, so in the previous example,
            they would each receive the points for 3rd. "average" averages the points, so they would
            each receive the points for 4th place.

    Returns:
        tuple[frozenset[str],...]: tiebroken ranking
    """
    if tiebreak == "random":
        new_ranking = tuple(
            frozenset({c}) for c in random.sample(list(r_set), k=len(r_set))
        )
    elif (tiebreak == "first_place" or tiebreak == "borda") and profile:
        if tiebreak == "borda":
            tiebreak_scores = borda_scores(
                profile, tie_convention=scoring_tie_convention
            )
        else:
            tiebreak_scores = first_place_votes(
                profile, tie_convention=scoring_tie_convention
            )
        tiebreak_scores = {
            c: Fraction(score) for c, score in tiebreak_scores.items() if c in r_set
        }
        new_ranking = score_dict_to_ranking(tiebreak_scores)

    elif not profile:
        raise ValueError("Method of tiebreak requires profile.")
    else:
        raise ValueError("Invalid tiebreak code was provided")

    if any(len(s) > 1 for s in new_ranking):
        print("Initial tiebreak was unsuccessful, performing random tiebreak")
        new_ranking, _ = tiebroken_ranking(
            new_ranking, profile=profile, tiebreak="random"
        )

    return new_ranking


def tiebroken_ranking(
    ranking: tuple[frozenset[str], ...],
    profile: Optional[PreferenceProfile] = None,
    tiebreak: str = "random",
) -> tuple[
    tuple[frozenset[str], ...], dict[frozenset[str], tuple[frozenset[str], ...]]
]:
    """
    Breaks ties in a list-of-sets ranking according to a given scheme.

    Args:
        ranking (list[set[str]]): A list-of-set ranking of candidates.
        profile (PreferenceProfile, optional): Profile used to break ties in first-place votes or
            Borda setting. Defaults to None, which implies a random tiebreak.
        tiebreak (str, optional): Method of tiebreak, currently supports 'random', 'borda',
            'first_place'. Defaults to random.

    Returns:
        tuple[tuple[frozenset[str], ...], dict[frozenset[str], tuple[frozenset[str],...]]]:
            The first entry of the tuple is a list-of-set ranking of candidates (broken down to one
            candidate sets). The second entry is a dictionary that maps tied sets to their
            resolution.
    """
    new_ranking: list[frozenset[str]] = [frozenset()] * len(
        [c for s in ranking for c in s]
    )

    i = 0
    tied_dict = {}
    for s in ranking:
        if len(s) > 1:
            tiebroken = tiebreak_set(s, profile, tiebreak)
            tied_dict[s] = tiebroken
        else:
            tiebroken = (s,)
        new_ranking[i : (i + len(tiebroken))] = tiebroken
        i += len(tiebroken)

    return (tuple(new_ranking), tied_dict)


def score_dict_to_ranking(
    score_dict: Union[dict[str, Fraction], dict[str, float]], sort_high_low: bool = True
) -> tuple[frozenset[str], ...]:
    """
    Sorts candidates into a tuple of frozensets ranking based on a scoring dictionary.

    Args:
        score_dict (Union[dict[str, Fraction],dict[str, float]]): Dictionary between candidates
            and their score.
        sort_high_low (bool, optional): How to sort candidates based on scores. True sorts
            from high to low. Defaults to True.


    Returns:
        tuple[frozenset[str],...]: Candidate rankings in a list-of-sets form.
    """

    score_to_cand: dict[Union[float, Fraction], list[str]] = {
        s: [] for s in score_dict.values()
    }
    for c, score in score_dict.items():
        score_to_cand[score].append(c)

    if len(score_to_cand) > 0:
        return tuple(
            [
                frozenset(c_list)
                for _, c_list in sorted(
                    score_to_cand.items(), key=lambda x: x[0], reverse=sort_high_low
                )
            ]
        )
    else:
        return (frozenset(),)


def elect_cands_from_set_ranking(
    ranking: tuple[frozenset[str], ...],
    m: int,
    profile: Optional[PreferenceProfile] = None,
    tiebreak: Optional[str] = None,
) -> tuple[
    tuple[frozenset[str], ...],
    tuple[frozenset[str], ...],
    Optional[tuple[frozenset[str], tuple[frozenset[str], ...]]],
]:
    """
    Given a ranking, elect the top m candidates in the ranking.
    If a tie set overlaps the desired number of seats, it breaks the tie with the provided
    method or raises a ValueError if tiebreak is set to None.
    Returns a tuple of elected candidates, remaining candidates, and a tuple whose first entry
    is a tie set and whose second entry is the resolution of the tie.

    Args:
        ranking (tuple[frozenset[str],...]): A list-of-set ranking of candidates.
        m (int): Number of seats to elect.
        profile (PreferenceProfile, optional): Profile used to break ties in first-place votes or
            Borda setting. Defaults to None, which implies a random tiebreak.
        tiebreak (str, optional): Method of tiebreak, currently supports 'random', 'borda',
            'first_place'. Defaults to None, which does not break ties.

    Returns:
        tuple[tuple[frozenset[str]]], list[tuple[frozenset[str]],
            Optional[tuple[frozenset[str], tuple[frozenset[str], ...]]]:
            A list-of-sets of elected candidates, a list-of-sets of remaining candidates,
            and a tuple whose first entry is a tie set and whose second entry is the resolution of
            the tie. If no ties were broken, the tuple returns None.
    """
    if m < 1:
        raise ValueError("m must be strictly positive")

    # if there are more seats than candidates
    if m > len([c for s in ranking for c in s]):
        raise ValueError("m must be no more than the number of candidates.")

    num_elected = 0
    elected = []
    i = 0
    tiebreak_ranking = None

    while num_elected < m:
        elected.append(ranking[i])
        num_elected += len(ranking[i])
        if num_elected > m:
            if not tiebreak:
                raise ValueError(
                    "Cannot elect correct number of candidates without breaking ties."
                )
            else:
                elected.pop(-1)
                num_elected -= len(ranking[i])
                tiebroken_ranking = tiebreak_set(ranking[i], profile, tiebreak)
                elected += tiebroken_ranking[: (m - num_elected)]
                remaining = list(tiebroken_ranking[(m - num_elected) :])
                if i < len(ranking):
                    remaining += list(ranking[(i + 1) :])

                return (
                    tuple(elected),
                    tuple(remaining),
                    (ranking[i], tiebroken_ranking),
                )

        i += 1

    return (tuple(elected), ranking[i:], tiebreak_ranking)


def expand_tied_ballot(ballot: Ballot) -> list[Ballot]:
    """
    Fix tie(s) in a ballot by returning all possible permutations of the tie(s), and divide the
    weight of the original ballot equally among the new ballots.

    Args:
        ballot (Ballot): Ballot to expand tie sets on.

    Returns:
        list[Ballot]: All possible permutations of the tie(s).

    """
    if not ballot.ranking:
        raise TypeError("Ballot must have ranking.")
    if all(len(s) == 1 for s in ballot.ranking):
        return [ballot]

    else:
        for i, s in enumerate(ballot.ranking):
            if len(s) > 1:
                new_ballots = [
                    Ballot(
                        weight=ballot.weight / math.factorial(len(s)),
                        id=ballot.id,
                        voter_set=ballot.voter_set,
                        ranking=tuple(ballot.ranking[:i])
                        + tuple([frozenset({c}) for c in order])
                        + tuple(ballot.ranking[(i + 1) :]),
                    )
                    for order in permutations(s)
                ]

                return [b for new_b in new_ballots for b in expand_tied_ballot(new_b)]

        assert False  # mypy


def resolve_profile_ties(profile: PreferenceProfile) -> PreferenceProfile:
    """
    Takes in a PeferenceProfile with potential ties in ballots. Replaces
    ballots with ties with fractionally weighted ballots corresponding to
    all permutations of the tied ranking. Automatically condenses the ballots.

    Args:
        profile (PreferenceProfile): Input profile with potentially tied rankings.

    Returns:
        PreferenceProfile: A PreferenceProfile with resolved ties.
    """

    new_ballots = tuple(
        [b for ballot in profile.ballots for b in expand_tied_ballot(ballot)]
    )
    return PreferenceProfile(ballots=new_ballots).condense_ballots()


def score_profile_from_ballot_scores(
    profile: PreferenceProfile,
    to_float: bool = False,
) -> Union[dict[str, Fraction], dict[str, float]]:
    """
    Score the candidates based on the ``scores`` parameter of the ballots.
    All ballots must have a ``scores`` parameter; note that a ``scores`` dictionary
    with no non-zero scores will raise the same error.

    Args:
        profile (PreferenceProfile): Profile to score.
        to_float (bool, optional): If True, compute scores as floats instead of Fractions.
            Defaults to False.

    Returns:
        Union[dict[str, Fraction], dict[str, float]]:
            Dictionary mapping candidates to scores.
    """
    scores = {c: Fraction(0) for c in profile.candidates}
    for ballot in profile.ballots:
        if not ballot.scores:
            raise TypeError(f"Ballot {ballot} has no scores.")
        else:
            for c, score in ballot.scores.items():
                scores[c] += score * ballot.weight

    if to_float:
        return {c: float(v) for c, v in scores.items()}
    return scores


def ballot_lengths(
    profile: PreferenceProfile, to_float: bool = False
) -> Union[dict[int, float], dict[int, Fraction]]:
    """
    Compute the frequency of ballot lengths in the profile.
    Includes all lengths from 1 to ``max_ballot_length`` as keys.
    Ballots must have rankings.

    Args:
        profile (PreferenceProfile): Profile to compute ballot lengths.
        to_float (bool): If True, compute mention as floats instead of Fractions.
            Defaults to False.

    Returns:
        Union[dict[int, float], dict[int, Fraction]]: Dictionary of ballot length frequency.

    Raises:
        TypeError: All ballots must have rankings.
    """

    ballot_lengths = {i: Fraction(0) for i in range(1, profile.max_ballot_length + 1)}

    for ballot in profile.ballots:
        if not ballot.ranking:
            raise TypeError("All ballots must have rankings.")

        length = len(ballot.ranking)
        ballot_lengths[length] += ballot.weight

    if to_float:
        return {length: float(f) for length, f in ballot_lengths.items()}

    return ballot_lengths



from .election_state import ElectionState  # noqa
from ..models import Election  # noqa
from .transfers import fractional_transfer, random_transfer  # noqa
from .election_types import (  # noqa
    RankingElection,
    Plurality,
    SNTV,
    Borda,
    STV,
    IRV,
    SequentialRCV,
    Alaska,
    DominatingSets,
    CondoBorda,
    TopTwo,
    GeneralRating,
    Rating,
    Limited,
    Cumulative,
    Approval,
    BlocPlurality,
    PluralityVeto,
    RandomDictator,
    BoostedRandomDictator,
)



from .mds import plot_MDS, compute_MDS  # noqa
from .bar_plot import multi_bar_plot, bar_plot  # noqa
from .profiles import (
    multi_profile_bar_plot,  # noqa
    multi_profile_mentions_plot,  # noqa
    multi_profile_ballot_lengths_plot,  # noqa
    multi_profile_borda_plot,  # noqa
    multi_profile_fpv_plot,  # noqa
    profile_bar_plot,  # noqa
    profile_borda_plot,  # noqa
    profile_ballot_lengths_plot,  # noqa
    profile_fpv_plot,  # noqa
    profile_mentions_plot,  # noqa
)



from fractions import Fraction
from pydantic.dataclasses import dataclass
from pydantic import ConfigDict, field_validator
from typing import Optional, Union


@dataclass(frozen=True, config=ConfigDict(arbitrary_types_allowed=True))
class Ballot:
    """
    Ballot class, contains ranking and assigned weight.

    Args:
        ranking (tuple[frozenset, ...], optional): Tuple of candidate ranking. Entry i of the tuple
            is a frozenset of candidates ranked in position i. Defaults to None.
        weight (Fraction, optional): Weight assigned to a given ballot. Defaults to 1.
            Can be input as int, float, or Fraction but will be converted to Fraction.
        voter_set (set[str], optional): Set of voters who cast the ballot. Defaults to None.
        id (str, optional): Ballot ID. Defaults to None.
        scores (dict[str, Fraction], optional): Scores for individual candidates. Defaults to None.
            Values can be input as int, float, or Fraction but will be converted to Fraction.
            Only retains non-zero scores.

    Attributes:
        ranking (tuple[frozenset, ...]): Tuple of candidate ranking. Entry i of the tuple is a
            frozenset of candidates ranked in position i.
        weight (Fraction): Weight assigned to a given ballot. Defaults to 1.
        voter_set (set[str], optional): Set of voters who cast the ballot. Defaults to None.
        id (str, optional): Ballot ID. Defaults to None.
        scores (dict[str, Fraction], optional): Scores for individual candidates. Defaults to None.
    """

    ranking: Optional[tuple[frozenset, ...]] = None
    weight: Fraction = Fraction(1, 1)
    voter_set: Optional[set[str]] = None
    id: Optional[str] = None
    scores: Optional[dict[str, Fraction]] = None

    @field_validator("weight", mode="before")
    @classmethod
    def convert_weight_to_fraction(cls, weight: Union[float, Fraction]) -> Fraction:
        if not isinstance(weight, Fraction):
            weight = Fraction(weight).limit_denominator()
        return weight

    @field_validator("scores", mode="before")
    @classmethod
    def convert_scores_to_fraction(
        cls, scores: Optional[dict[str, Union[float, Fraction]]]
    ) -> Optional[dict[str, Fraction]]:
        if scores:
            if any(
                not (
                    isinstance(s, float)
                    or isinstance(s, Fraction)
                    or isinstance(s, int)
                )
                for s in scores.values()
            ):
                raise TypeError("Score values must be numeric.")

            return {
                c: Fraction(s).limit_denominator() for c, s in scores.items() if s != 0
            }
        else:
            return None

    def __eq__(self, other):
        # Check type
        if not isinstance(other, Ballot):
            return False

        # Check id
        if self.id is not None:
            if self.id != other.id:
                return False

        # Check ranking
        if self.ranking != other.ranking:
            return False

        # Check weight
        if self.weight != other.weight:
            return False

        # Check voters
        if self.voter_set is not None:
            if self.voter_set != other.voter_set:
                return False

        # Check scores
        if self.scores is not None:
            if self.scores != other.scores:
                return False
        return True

    def __hash__(self):
        return hash(self.ranking)

    def __str__(self):
        weight_str = f"Weight: {self.weight}"

        if self.ranking:
            ranking_str = "Ranking\n"
            for i, s in enumerate(self.ranking):
                # display number and candidates
                ranking_str += f"{i+1}.) "
                for c in s:
                    ranking_str += f"{c}, "

                # if tie
                if len(s) > 1:
                    ranking_str += "(tie)"
                ranking_str += "\n"
        else:
            ranking_str = ""

        if self.scores:
            score_str = "Scores\n"
            for c, score in self.scores.items():
                score_str += f"{c}: {float(score):.2f}\n"
        else:
            score_str = ""
        return ranking_str + score_str + weight_str

    __repr__ = __str__



from .candidate import (
    comention,  # noqa
    comention_above,  # noqa
    comentions_matrix,  # noqa
    boost_prob,  # noqa
    boost_matrix,  # noqa
    candidate_distance,  # noqa
    candidate_distance_matrix,  # noqa
)

from .heatmap import matrix_heatmap  # noqa



from .abstract_ranking import RankingElection
from ...transfers import fractional_transfer
from ....pref_profile import PreferenceProfile
from ...election_state import ElectionState
from ....ballot import Ballot
from ....utils import (
    remove_cand,
    first_place_votes,
    ballots_by_first_cand,
    tiebreak_set,
    elect_cands_from_set_ranking,
    score_dict_to_ranking,
)
from typing import Optional, Callable, Union, Literal
from fractions import Fraction
from functools import partial


class STV(RankingElection):
    """
    STV elections. Currently implements simultaneous election. All ballots must have no
    ties.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        m (int, optional): Number of seats to be elected. Defaults to 1.
        transfer (Callable[[str, Union[Fraction, float], Union[tuple[Ballot], list[Ballot]], int], tuple[Ballot,...]], optional):
        Transfer method. Defaults to fractional transfer.
            Function signature is elected candidate, their number of first-place votes, the list of
            ballots with them ranked first, and the threshold value. Returns the list of ballots
            after transfer.
        quota (str, optional): Formula to calculate quota. Accepts "droop" or "hare".
            Defaults to "droop".
        simultaneous (bool, optional): True if all candidates who cross threshold in a round are
            elected simultaneously, False if only the candidate with highest first-place votes
            who crosses the threshold is elected in a round. Defaults to True.
        tiebreak (str, optional): Method to be used if a tiebreak is needed. Accepts
            'borda' and 'random'. Defaults to None, in which case a ValueError is raised if
            a tiebreak is needed.
        fpv_tie_convention (Literal["high", "average", "low"], optional): How to award points
            for tied first place votes. Defaults to "average", where if n candidates are tied for
            first, each receives 1/n points. "high" would award them each one point, and "low" 0.
            Only used by ``score_function`` parameter.

    """  # noqa

    def __init__(
        self,
        profile: PreferenceProfile,
        m: int = 1,
        transfer: Callable[
            [str, Union[Fraction, float], Union[tuple[Ballot], list[Ballot]], int],
            tuple[Ballot, ...],
        ] = fractional_transfer,
        quota: str = "droop",
        simultaneous: bool = True,
        tiebreak: Optional[str] = None,
        fpv_tie_convention: Literal["high", "low", "average"] = "average",
    ):
        self._stv_validate_profile(profile)

        if m <= 0 or m > len(profile.candidates):
            raise ValueError(
                "m must be non-negative and less than or equal to the number of candidates."
            )

        self.m = m
        self.transfer = transfer
        self.quota = quota
        # Set to 0 initially so that first call to `get_threshold` returns the
        # proper threshold.
        self.threshold = 0
        self.threshold = self.get_threshold(profile.total_ballot_wt)
        self.simultaneous = simultaneous
        self.tiebreak = tiebreak
        super().__init__(
            profile,
            score_function=partial(
                first_place_votes, tie_convention=fpv_tie_convention
            ),
            sort_high_low=True,
        )

    def _stv_validate_profile(self, profile: PreferenceProfile):
        """
        Validate that each ballot has a ranking, and that there are no ties in ballots.
        """

        for ballot in profile.ballots:
            if not ballot.ranking:
                raise TypeError("Ballots must have rankings.")
            if len(ballot.ranking) == 0:
                raise TypeError("All ballots must have rankings.")
            elif any(len(s) > 1 for s in ballot.ranking):
                raise TypeError(f"Ballot {ballot} contains a tied ranking.")

    def get_threshold(self, total_ballot_wt: Fraction) -> int:
        """
        Calculates threshold required for election.

        Args:
            total_ballot_wt (Fraction): Total weight of ballots to compute threshold.
        Returns:
            int: Value of the threshold.
        """
        if self.threshold == 0:
            if self.quota == "droop":
                return int(total_ballot_wt / (self.m + 1) + 1)  # takes floor
            elif self.quota == "hare":
                return int(total_ballot_wt / self.m)  # takes floor
            else:
                raise ValueError("Misspelled or unknown quota type.")
        else:
            return self.threshold

    def _is_finished(self):
        elected_cands = [c for s in self.get_elected() for c in s]

        if len(elected_cands) == self.m:
            return True
        return False

    def _simultaneous_elect_step(
        self, profile: PreferenceProfile, prev_state: ElectionState
    ) -> tuple[tuple[frozenset[str], ...], PreferenceProfile]:
        """
        Run one step of an election from the given profile and previous state.
        Used for simultaneous STV election if candidates cross threshold.

        Args:
            profile (PreferenceProfile): Profile of ballots.
            prev_state (ElectionState): The previous ElectionState.

        Returns:
            tuple[tuple[frozenset[str],...], PreferenceProfile]:
                A tuple whose first entry is the elected candidates, ranked by first-place votes,
                and whose second entry is the profile of ballots after transfers.
        """
        ranking_by_fpv = prev_state.remaining

        elected = []
        for s in ranking_by_fpv:
            c = list(s)[0]  # all cands in set have same score
            if prev_state.scores[c] >= self.threshold:
                elected.append(s)

            # since ranking is ordered by fpv, once below threshold we are done
            else:
                break

        ballots_by_fpv = ballots_by_first_cand(profile)
        new_ballots = [Ballot()] * profile.num_ballots
        ballot_index = 0

        for s in elected:
            for candidate in s:
                transfer_ballots = self.transfer(
                    candidate,
                    prev_state.scores[candidate],
                    ballots_by_fpv[candidate],
                    self.threshold,
                )
                new_ballots[ballot_index : (ballot_index + len(transfer_ballots))] = (
                    transfer_ballots
                )
                ballot_index += len(transfer_ballots)

        for candidate in set([c for s in ranking_by_fpv for c in s]).difference(
            [c for s in elected for c in s]
        ):
            transfer_ballots = tuple(ballots_by_fpv[candidate])
            new_ballots[ballot_index : (ballot_index + len(transfer_ballots))] = (
                transfer_ballots
            )
            ballot_index += len(transfer_ballots)

        cleaned_ballots = remove_cand(
            [c for s in elected for c in s],
            tuple([b for b in new_ballots if b.ranking]),
        )

        remaining_cands = set(profile.candidates).difference(
            [c for s in elected for c in s]
        )
        new_profile = PreferenceProfile(
            ballots=cleaned_ballots, candidates=tuple(remaining_cands)
        )
        return (tuple(elected), new_profile)

    def _single_elect_step(
        self, profile: PreferenceProfile, prev_state: ElectionState
    ) -> tuple[
        tuple[frozenset[str], ...],
        dict[frozenset[str], tuple[frozenset[str], ...]],
        PreferenceProfile,
    ]:
        """
        Run one step of an election from the given profile and previous state.
        Used for one-by-one STV election if candidates cross threshold.

        Args:
            profile (PreferenceProfile): Profile of ballots.
            prev_state (ElectionState): The previous ElectionState.

        Returns:
            tuple[tuple[frozenset[str],...], dict[frozenset[str], tuple[frozenset[str],...]],
            PreferenceProfile]:
                A tuple whose first entry is the elected candidate, second is the tiebreak dict,
                and whose third entry is the profile of ballots after transfers.
        """
        ranking_by_fpv = prev_state.remaining

        elected, remaining, tiebreak = elect_cands_from_set_ranking(
            ranking_by_fpv, m=1, profile=profile, tiebreak=self.tiebreak
        )
        if tiebreak:
            tiebreaks = {tiebreak[0]: tiebreak[1]}
        else:
            tiebreaks = {}

        ballots_by_fpv = ballots_by_first_cand(profile)
        new_ballots = [Ballot()] * profile.num_ballots
        ballot_index = 0

        elected_c = list(elected[0])[0]

        transfer_ballots = self.transfer(
            elected_c,
            prev_state.scores[elected_c],
            ballots_by_fpv[elected_c],
            self.threshold,
        )
        new_ballots[ballot_index : (ballot_index + len(transfer_ballots))] = (
            transfer_ballots
        )
        ballot_index += len(transfer_ballots)

        for s in remaining:
            for candidate in s:
                transfer_ballots = tuple(ballots_by_fpv[candidate])
                new_ballots[ballot_index : (ballot_index + len(transfer_ballots))] = (
                    transfer_ballots
                )
                ballot_index += len(transfer_ballots)

        cleaned_ballots = remove_cand(
            elected_c, tuple([b for b in new_ballots if b.ranking])
        )

        remaining_cands = set(profile.candidates).difference(
            [c for s in elected for c in s]
        )
        new_profile = PreferenceProfile(
            ballots=cleaned_ballots, candidates=tuple(remaining_cands)
        )
        return elected, tiebreaks, new_profile

    def _run_step(
        self, profile: PreferenceProfile, prev_state: ElectionState, store_states=False
    ) -> PreferenceProfile:
        """
        Run one step of an election from the given profile and previous state.
        STV sets a threshold for first-place votes. If a candidate passes it, they are elected.
        We remove them from all ballots and transfer any surplus ballots to other candidates.
        If no one passes, we eliminate the lowest ranked candidate and reallocate their ballots.

        Can be run 1-by-1 or simultaneous, which determines what happens if multiple people cross
        threshold.

        Args:
            profile (PreferenceProfile): Profile of ballots.
            prev_state (ElectionState): The previous ElectionState.
            store_states (bool, optional): True if `self.election_states` should be updated with the
                ElectionState generated by this round. This should only be True when used by
                `self._run_election()`. Defaults to False.

        Returns:
            PreferenceProfile: The profile of ballots after the round is completed.
        """
        tiebreaks: dict[frozenset[str], tuple[frozenset[str], ...]] = {}

        above_thresh_cands = [
            c for c, score in prev_state.scores.items() if score >= self.threshold
        ]

        if len(above_thresh_cands) > 0:
            if self.simultaneous:
                elected, new_profile = self._simultaneous_elect_step(
                    profile, prev_state
                )

            else:
                elected, tiebreaks, new_profile = self._single_elect_step(
                    profile, prev_state
                )

            # no on eliminated in elect round
            eliminated: tuple[frozenset[str], ...] = (frozenset(),)

        # catches the possibility that we exhaust all ballots
        # without candidates reaching threshold
        elif len(profile.candidates) == self.m - len(
            [c for s in self.get_elected() for c in s]
        ):
            elected = prev_state.remaining
            eliminated = (frozenset(),)
            new_profile = PreferenceProfile()

        else:
            lowest_fpv_cands = prev_state.remaining[-1]

            if len(lowest_fpv_cands) > 1:
                tiebroken_ranking = tiebreak_set(
                    lowest_fpv_cands, self.get_profile(0), tiebreak="first_place"
                )
                tiebreaks = {lowest_fpv_cands: tiebroken_ranking}

                eliminated_cand = list(tiebroken_ranking[-1])[0]

            else:
                eliminated_cand = list(lowest_fpv_cands)[0]

            new_profile = remove_cand(eliminated_cand, profile)
            elected = (frozenset(),)
            eliminated = (frozenset([eliminated_cand]),)

        if store_states:
            if self.score_function:
                scores = self.score_function(new_profile)

            remaining = score_dict_to_ranking(scores)

            new_state = ElectionState(
                round_number=prev_state.round_number + 1,
                remaining=remaining,
                elected=elected,
                eliminated=eliminated,
                scores=scores,
                tiebreaks=tiebreaks,
            )

            self.election_states.append(new_state)

        return new_profile


class IRV(STV):
    """
    IRV (Instant-runoff voting) elections.  Elect 1 seat. All ballots must have no ties.
    Equivalent to STV for m = 1.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        quota (str, optional): Formula to calculate quota. Accepts "droop" or "hare".
            Defaults to "droop".
        tiebreak (str, optional): Method to be used if a tiebreak is needed. Accepts
            'borda' and 'random'. Defaults to None, in which case a ValueError is raised if
            a tiebreak is needed.

    """

    def __init__(
        self,
        profile: PreferenceProfile,
        quota: str = "droop",
        tiebreak: Optional[str] = None,
    ):
        super().__init__(profile, m=1, quota=quota, tiebreak=tiebreak)


class SequentialRCV(STV):
    """
    An STV election in which votes are not transferred after a candidate has reached threshold, or
    been elected. This system is actually used in parts of Utah.

    Args:
        profile (PreferenceProfile):   PreferenceProfile to run election on.
        m (int, optional): Number of seats to be elected. Defaults to 1.
        quota (str, optional): Formula to calculate quota. Accepts "droop" or "hare".
            Defaults to "droop".
        simultaneous (bool, optional): True if all candidates who cross threshold in a round are
            elected simultaneously, False if only the candidate with highest first-place votes
            who crosses the threshold is elected in a round. Defaults to True.
        tiebreak (str, optional): Method to be used if a tiebreak is needed. Accepts
            'borda' and 'random'. Defaults to None, in which case a ValueError is raised if
            a tiebreak is needed.

    """

    def __init__(
        self,
        profile: PreferenceProfile,
        m: int = 1,
        quota: str = "droop",
        simultaneous: bool = True,
        tiebreak: Optional[str] = None,
    ):
        super().__init__(
            profile,
            m=m,
            transfer=(
                lambda winner, fpv, ballots, threshold: remove_cand(
                    winner, tuple(ballots)
                )
            ),
            quota=quota,
            simultaneous=simultaneous,
            tiebreak=tiebreak,
        )



from .profile_bar_plot import (
    profile_bar_plot,  # noqa
    profile_borda_plot,  # noqa
    profile_mentions_plot,  # noqa
    profile_fpv_plot,  # noqa
    profile_ballot_lengths_plot,  # noqa
)

from .multi_profile_bar_plot import (
    multi_profile_bar_plot,  # noqa
    multi_profile_ballot_lengths_plot,  # noqa
    multi_profile_borda_plot,  # noqa
    multi_profile_fpv_plot,  # noqa
    multi_profile_mentions_plot,  # noqa
)



from .ranking import (  # noqa
    RankingElection,
    Plurality,
    SNTV,
    Borda,
    STV,
    IRV,
    Alaska,
    DominatingSets,
    SequentialRCV,
    CondoBorda,
    TopTwo,
    PluralityVeto,
    RandomDictator,
    BoostedRandomDictator,
)


from .scores import GeneralRating, Rating, Limited, Cumulative  # noqa
from .approval import Approval, BlocPlurality  # noqa



from __future__ import annotations
import csv
from fractions import Fraction
import pandas as pd
from pydantic import ConfigDict, field_validator, model_validator
from typing import Optional
from .ballot import Ballot
from pydantic.dataclasses import dataclass
from typing_extensions import Self
from dataclasses import field


@dataclass(frozen=True, config=ConfigDict(arbitrary_types_allowed=True))
class PreferenceProfile:
    """
    PreferenceProfile class, contains ballots and candidates for a given election.
    This is a frozen class, so you need to create a new PreferenceProfile any time
    you want to edit the ballots, candidates, etc.

    Args:
        ballots (tuple[Ballot], optional): Tuple of ``Ballot`` objects. Defaults to empty tuple.
        candidates (tuple[str], optional): Tuple of candidate strings. Defaults to empty tuple.
            If empty, computes this from any candidate listed on a ballot with positive weight.
        max_ballot_length (int, optional): The length of the longest allowable ballot, i.e., how
            many candidates are allowed to be ranked in an election. Defaults to longest observed
            ballot.

    Parameters:
        ballots (tuple[Ballot]): Tuple of ``Ballot`` objects.
        candidates (tuple[str]): Tuple of candidate strings.
        max_ballot_length (int): The length of the longest allowable ballot, i.e., how
            many candidates are allowed to be ranked in an election.
        df (pandas.DataFrame): Data frame view of the ballots.
        candidates_cast (tuple[str]): Tuple of candidates who appear on any ballot with positive
            weight, either in the ranking or in the score dictionary.
        total_ballot_wt (Fraction): Sum of ballot weights.
        num_ballots (int): Length of ballot list.

    """

    ballots: tuple[Ballot, ...] = field(default_factory=tuple)
    candidates: tuple[str, ...] = field(default_factory=tuple)
    max_ballot_length: int = 0
    df: pd.DataFrame = field(default_factory=pd.DataFrame)
    candidates_cast: tuple[str, ...] = field(default_factory=tuple)
    num_ballots: int = 0
    total_ballot_wt: Fraction = Fraction(0)

    @field_validator("candidates")
    @classmethod
    def cands_must_be_unique(
        cls, candidates: Optional[tuple[str, ...]]
    ) -> Optional[tuple[str, ...]]:
        if candidates:
            if not len(set(candidates)) == len(candidates):
                raise ValueError("All candidates must be unique.")
        return candidates

    @model_validator(mode="after")
    def find_candidates_cast(self) -> Self:
        candidates_cast: set = set()
        for ballot in self.ballots:
            if ballot.weight > 0:
                if ballot.ranking:
                    candidates_cast.update(*ballot.ranking)
                if ballot.scores:
                    candidates_cast.update(ballot.scores.keys())

        object.__setattr__(self, "candidates_cast", tuple(candidates_cast))
        if not self.candidates:
            object.__setattr__(self, "candidates", tuple(candidates_cast))
        return self

    @model_validator(mode="after")
    def check_ballot_length(self) -> Self:
        if self.max_ballot_length == 0:
            for ballot in self.ballots:
                ballot_len = len(ballot.ranking) if ballot.ranking else 0

                if ballot_len > self.max_ballot_length:
                    object.__setattr__(self, "max_ballot_length", ballot_len)

        else:
            for ballot in self.ballots:
                if ballot.ranking:
                    if len(ballot.ranking) > self.max_ballot_length:
                        raise ValueError(
                            "Profile contains a ballot with a ranking that is too long."
                        )

        return self

    @model_validator(mode="after")
    def find_num_ballots(self) -> Self:
        object.__setattr__(self, "num_ballots", len(self.ballots))
        return self

    @model_validator(mode="after")
    def find_total_ballot_wt(self) -> Self:
        total_ballot_wt = Fraction(0)
        for ballot in self.ballots:
            total_ballot_wt += ballot.weight

        object.__setattr__(self, "total_ballot_wt", total_ballot_wt)

        return self

    @model_validator(mode="after")
    def create_df(self) -> Self:
        weights = []
        rankings = []
        scores: list[tuple[str, ...]] = []
        for ballot in self.ballots:
            part = []
            if ballot.ranking:
                for ranking in ballot.ranking:
                    if len(ranking) == 1:
                        part.append(list(ranking)[0])

                    else:
                        part.append(f"{set(ranking)} (Tie)")
            score_list: list[str] = []
            if ballot.scores:
                for c, score in ballot.scores.items():
                    score_list.append(f"{c}:{float(score):.2f}")

            rankings.append(tuple(part))
            weights.append(ballot.weight)
            scores.append(tuple(score_list))

        df = pd.DataFrame({"Ranking": rankings, "Scores": scores, "Weight": weights})

        try:
            df["Percent"] = df["Weight"] / df["Weight"].sum()
        except ZeroDivisionError:
            df["Percent"] = 0.0

        def format_as_percent(frac):
            return f"{float(frac):.2%}"

        df["Percent"] = df["Percent"].apply(format_as_percent)
        object.__setattr__(self, "df", df)
        return self

    def to_ballot_dict(self, standardize: bool = False) -> dict[Ballot, Fraction]:
        """
        Converts profile to dictionary with keys = ballots and
        values = corresponding total weights.

        Args:
            standardize (bool, optional): If True, divides the weight of each ballot by the total
                weight. Defaults to False.

        Returns:
            dict[Ballot, Fraction]:
                A dictionary with ballots (keys) and corresponding total weights (values).
        """
        tot_weight = self.total_ballot_wt
        di: dict = {}
        for ballot in self.ballots:
            weightless_ballot = Ballot(ranking=ballot.ranking, scores=ballot.scores)
            if standardize:
                weight = ballot.weight / tot_weight
            else:
                weight = ballot.weight
            if weightless_ballot not in di.keys():
                di[weightless_ballot] = weight
            else:
                di[weightless_ballot] += weight
        return di

    def to_ranking_dict(
        self, standardize: bool = False
    ) -> dict[tuple[frozenset[str], ...], Fraction]:
        """
        Converts profile to dictionary with keys = rankings and
        values = corresponding total weights.

        Args:
            standardize (bool, optional): If True, divides the weight of each ballot by the total
                weight. Defaults to False.

        Returns:
            dict[tuple[frozenset[str],...], Fraction]:
                A dictionary with rankings (keys) and corresponding total weights (values).
        """
        tot_weight = self.total_ballot_wt
        di: dict = {}
        for ballot in self.ballots:
            ranking = ballot.ranking

            if not ranking:
                ranking = (frozenset(),)
            if standardize:
                weight = ballot.weight / tot_weight
            else:
                weight = ballot.weight

            if ranking not in di.keys():
                di[ranking] = weight
            else:
                di[ranking] += weight
        return di

    def to_scores_dict(
        self, standardize: bool = False
    ) -> dict[tuple[str, Fraction], Fraction]:
        """
        Converts profile to dictionary with keys = scores and
        values = corresponding total weights.

        Args:
            standardize (bool, optional): If True, divides the weight of each ballot by the total
                weight. Defaults to False.

        Returns:
            dict[tuple[str, Fraction], Fraction]:
                A dictionary with scores (keys) and corresponding total weights (values).
        """
        tot_weight = self.total_ballot_wt
        di: dict = {}
        for ballot in self.ballots:
            if ballot.scores:
                scores = tuple([(c, score) for c, score in ballot.scores.items()])
            else:
                scores = tuple()
            if standardize:
                weight = ballot.weight / tot_weight
            else:
                weight = ballot.weight

            if scores not in di.keys():
                di[scores] = weight
            else:
                di[scores] += weight
        return di

    def to_csv(self, fpath: str):
        """
        Saves PreferenceProfile to CSV.

        Args:
            fpath (str): Path to the saved csv.
        """
        with open(fpath, "w", newline="") as csvfile:
            fieldnames = ["weight", "ranking", "scores"]
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            writer.writeheader()
            for ballot in self.ballots:
                if ballot.ranking:
                    ranking = tuple([set(s) for s in ballot.ranking])
                else:
                    ranking = tuple()

                if ballot.scores:
                    scores = tuple(
                        [(c, float(score)) for c, score in ballot.scores.items()]
                    )
                else:
                    scores = tuple()
                writer.writerow(
                    {
                        "weight": float(ballot.weight),
                        "ranking": ranking,
                        "scores": scores,
                    }
                )

    def head(
        self,
        n: int,
        sort_by_weight: Optional[bool] = True,
        percents: Optional[bool] = False,
        totals: Optional[bool] = False,
    ) -> pd.DataFrame:
        """
        Displays top-n ballots in profile.

        Args:
            n (int): Number of ballots to view.
            sort_by_weight (bool, optional): If True, rank ballot from most to least votes.
                Defaults to True.
            percents (bool, optional): If True, show voter share for a given ballot.
                Defaults to False.
            totals (bool, optional): If True, show total values for Percent and Weight.
                Defaults to False.

        Returns:
            pandas.DataFrame: A dataframe with top-n ballots.
        """
        if sort_by_weight:
            df = (
                self.df.sort_values(by="Weight", ascending=False)
                .head(n)
                .reset_index(drop=True)
            )
        else:
            df = self.df.head(n).reset_index(drop=True)

        if totals:
            df = self._sum_row(df)

        if not percents:
            return df.drop(columns="Percent")

        return df

    def tail(
        self,
        n: int,
        sort_by_weight: Optional[bool] = True,
        percents: Optional[bool] = False,
        totals: Optional[bool] = False,
    ) -> pd.DataFrame:
        """
        Displays bottom-n ballots in profile.

        Args:
            n (int): Number of ballots to view.
            sort_by_weight (bool, optional): If True, rank ballot from least to most votes.
                Defaults to True.
            percents (bool, optional): If True, show voter share for a given ballot.
                Defaults to False.
            totals (bool, optional): If True, show total values for Percent and Weight.
                Defaults to False.

        Returns:
            pandas.DataFrame: A data frame with bottom-n ballots.
        """
        if sort_by_weight:
            df = self.df.sort_values(by="Weight", ascending=True)
            df["New Index"] = [x for x in range(len(self.df) - 1, -1, -1)]
            df = df.set_index("New Index").head(n)
            df.index.name = None

        else:
            df = self.df.iloc[::-1].head(n)

        if totals:
            df = self._sum_row(df)

        if not percents:
            return df.drop(columns="Percent")

        return df

    def __str__(self) -> str:
        # Displays top 15 cast ballots or entire profile

        if len(self.df) < 15:
            return self.head(n=len(self.df), sort_by_weight=True).to_string(
                index=False, justify="justify"
            )

        print(
            f"PreferenceProfile too long, only showing 15 out of {len(self.df) } rows."
        )
        return self.head(n=15, sort_by_weight=True).to_string(
            index=False, justify="justify"
        )

    # set repr to print outputs
    __repr__ = __str__

    def condense_ballots(self) -> PreferenceProfile:
        """
        Groups ballots by rankings and scores and updates weights.

        Returns:
            PreferenceProfile: A PreferenceProfile object with condensed ballot list.
        """
        weight_accumulator = {}

        # weightless allows for id of ballots with matching ranking/scores
        for ballot in self.ballots:
            weightless_ballot = (
                Ballot(ranking=ballot.ranking, weight=Fraction(0), scores=ballot.scores)
                if ballot.scores
                else Ballot(ranking=ballot.ranking, weight=Fraction(0))
            )
            if weightless_ballot not in weight_accumulator:
                weight_accumulator[weightless_ballot] = Fraction(0)

            weight_accumulator[weightless_ballot] += ballot.weight

        new_ballot_list = [Ballot()] * len(weight_accumulator)
        i = 0
        for ballot, weight in weight_accumulator.items():
            if ballot.scores:
                new_ballot_list[i] = Ballot(
                    ranking=ballot.ranking, scores=ballot.scores, weight=weight
                )
            else:
                new_ballot_list[i] = Ballot(ranking=ballot.ranking, weight=weight)

            i += 1

        condensed_profile = PreferenceProfile(
            ballots=tuple(new_ballot_list), candidates=self.candidates
        )
        return condensed_profile

    def __eq__(self, other):
        if not isinstance(other, PreferenceProfile):
            return False
        pp_1 = self.condense_ballots()
        pp_2 = other.condense_ballots()
        for b in pp_1.ballots:
            if b not in pp_2.ballots:
                return False
        for b in pp_2.ballots:
            if b not in pp_1.ballots:
                return False
        return True

    def _sum_row(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Computes sum total for weight and percent column
        """

        def format_as_float(percent_str):
            return float(percent_str.split("%")[0])

        sum_row = {
            "Ranking": "",
            "Scores": "",
            "Weight": f'{df["Weight"].sum()} out of {self.total_ballot_wt}',
            "Percent": f'{df["Percent"].apply(format_as_float).sum():.2f} out of 100%',
        }

        df.loc["Totals"] = sum_row  # type: ignore

        return df.fillna("")

    def __add__(self, other):
        """
        Add two PreferenceProfiles by combining their ballot lists.
        """
        if isinstance(other, PreferenceProfile):
            ballots = self.ballots + other.ballots
            pp = PreferenceProfile(ballots=ballots)
            pp.condense_ballots()
            return pp
        else:
            raise TypeError(
                "Unsupported operand type. Must be an instance of PreferenceProfile."
            )
